#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) since 2015 NTT DATA Corporation 
# Last update : 2015/09/15
#
from __future__ import unicode_literals

import argparse
import codecs
import docker
import logging.config
import os
import pprint
import re
import sys
import tarfile
import traceback

SCRIPT_NAME = os.path.splitext(os.path.basename(sys.argv[0]))[0]
DOCKER_API_VERSION = "1.18"
DEFAULT_HOST = "127.0.0.1"
DEFAULT_PORT = "2375"
SEP_SUBST = "--"    # パス区切り文字の代替文字列

# 標準入出力をUTF-8で行う
sys.stdin = codecs.getreader("utf-8")(sys.stdin)
sys.stdout = codecs.getwriter("utf-8")(sys.stdout)

# カレントディレクトリをスクリプトのディレクトリへ移動
# loggingを初期化(logging.iniファイルがある場合)
os.chdir(os.path.dirname(os.path.abspath(sys.argv[0])))
if os.path.exists("logging.ini"): logging.config.fileConfig("logging.ini")
logger = logging.getLogger(SCRIPT_NAME)

class LocalException(Exception):
    pass

# 開発/デバッグ用
def dumpvar(var):
    print pprint.pformat(var).decode("utf-8")

#---------------------------------------
# メイン
#---------------------------------------
def main():
    logger.info(SCRIPT_NAME)

    # 引数解析
    parser = argparse.ArgumentParser(
            formatter_class=argparse.RawTextHelpFormatter,
            description="Dockerのイベント情報を取得します。")

    parser.add_argument(
            "-i", dest="ip_addr", metavar="IP_ADDRESS",
            default=DEFAULT_HOST,
            help="Docker Remote APIの接続先IPアドレス。(default: '%s')" % DEFAULT_HOST)
    parser.add_argument(
            "-p", dest="port", metavar="DOCKER_PORT",
            default=DEFAULT_PORT,
            help="Docker Remote APIの接続先ポート。(default: '%s')" % DEFAULT_PORT)
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument(
            "-c", dest="cid", metavar="CONTAINER_ID", 
            help="コンテナのIDパターンを指定します。")
    group.add_argument(
            "-n", dest="name", metavar="CONTAINER_NAME",
            help="コンテナの名称パターンを指定します。")
    parser.add_argument(
            dest="src_paths", metavar="SOURCE", nargs="+",
            help="コンテナからダウンロードするファイル/ディレクトリを指定します。")
    parser.add_argument(
            "dest_path", metavar="DESTINATION",
            help="ダウンロード先ディレクトリを指定します。")


    args = parser.parse_args()

    # Docker接続先URLの構築
    args.url = "http://%s:%s" % (args.ip_addr, args.port)

    # 引数を記録
    logger.info("args:%s", args)

    # ダウンロード先のチェック
    if not os.path.isdir(args.dest_path):
        raise LocalException("ダウンロード先ディレクトリが存在しません。")

    # コンテナIDのリストを作成する
    dc = docker.Client(base_url=args.url, version=DOCKER_API_VERSION)
    containers = []
    cid2name = {}
    if args.cid :
        pa = re.compile(('%s.*' % args.cid[:len(args.cid) - 1]) if args.cid.endswith('*') else args.cid)
        def find_container_by_id(c):
            if pa.match(c["Id"]):
                containers.append(c["Id"])
                cid2name[c["Id"]] = ["Names"][0][1:]
        map(find_container_by_id, dc.containers(quiet=True))
    elif args.name :
        pa = re.compile(('%s.*' % args.name[:len(args.name) - 1]) if args.name.endswith('*') else args.name)
        def find_container_by_name(c):
            # 名前に含まれる '/' を除外するため、マッチングする文字列の 1 文字目をスキップ
            matched_names = [n[1:] for n in c["Names"] if pa.match(n[1:])]
            if len(matched_names) > 0:
                containers.append(c["Id"])
                cid2name[c["Id"]] = ["Names"][0][1:]
        map(find_container_by_name, dc.containers())
        
    if len(containers) == 0:
        raise LocalException("指定されたコンテナは存在しません。")

    # ファイル取得実行
    dc = docker.Client(base_url=args.url, version=DOCKER_API_VERSION)
    for cid in containers:
        cnm = cid2name[cid]
        for src in args.src_paths:
            # 出力ファイルパスを生成
            outfile_path = "%s%s%s_%s.tar" % (args.dest_path, os.sep, cnm, src.lstrip(os.sep).replace(os.sep, SEP_SUBST))
            print outfile_path
            logger.info(outfile_path)

            # ファイルを取得、保存
            try:
                with dc.copy(container=cid, resource=src) as res:
                    with open(outfile_path, "wb") as f:
                        f.write(res.read())
            except docker.errors.APIError as e:
                # 対象ディレクトリがない場合は処理をスキップ
                errmsg = str(e)
                if "Could not find the file" in errmsg:
                    logger.warn(errmsg)
                    print "[skip] %s 上に %s は存在しません。" % (cnm, src)
                    continue

            # tarに含まれるファイルのリストを出力
            with tarfile.open(name=outfile_path) as tar:
                tar.list(verbose=True)

#===============================================================================
if __name__ == "__main__":
    try:
        main()
        sys.exit(0)

    # 予期されている例外はメッセージのみ出力
    except LocalException as e:
        print e.message
        sys.exit(1)

    # 予期しない例外はスタックトレースまで出力
    except Exception as e:
        print e.message
        traceback.print_exc()
        sys.exit(1)
