#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) since 2015 NTT DATA Corporation 
# Last update : 2015/09/15
#
from __future__ import unicode_literals

import argparse
import codecs
import docker
import logging.config
import os

import sys
import traceback
import re

SCRIPT_NAME = os.path.splitext(os.path.basename(sys.argv[0]))[0]
DOCKER_API_VERSION = "1.18"
DEFAULT_HOST = "127.0.0.1"
DEFAULT_PORT = "2375"

# 標準入出力をUTF-8で行う
sys.stdin = codecs.getreader("utf-8")(sys.stdin)
sys.stdout = codecs.getwriter("utf-8")(sys.stdout)

# カレントディレクトリをスクリプトのディレクトリへ移動
# loggingを初期化(logging.iniファイルがある場合)
os.chdir(os.path.dirname(os.path.abspath(sys.argv[0])))
if os.path.exists("logging.ini"): logging.config.fileConfig("logging.ini")
logger = logging.getLogger(SCRIPT_NAME)

class LocalException(Exception):
    pass

#---------------------------------------
# メイン
#---------------------------------------
def main():
    logger.info(SCRIPT_NAME)

    # 引数解析
    parser = argparse.ArgumentParser(
            formatter_class=argparse.RawTextHelpFormatter,
            description="Docker コンテナ上で動作するプロセス数を取得します。")

    parser.add_argument(
            "-i", dest="ip_addr", metavar="IP_ADDRESS",
            default=DEFAULT_HOST,
            help="Docker Remote APIの接続先IPアドレス。(default: '%s')" % DEFAULT_HOST)
    parser.add_argument(
            "-p", dest="port", metavar="DOCKER_PORT",
            default=DEFAULT_PORT,
            help="Docker Remote APIの接続先ポート。(default: '%s')" % DEFAULT_PORT)
    group = parser.add_mutually_exclusive_group()
    group.add_argument(
            "-c", dest="cid", metavar="CONTAINER_ID", 
            help="コンテナのIDパターンを指定します。")
    group.add_argument(
            "-n", dest="name", metavar="CONTAINER_NAME",
            help="コンテナの名称パターンを指定します。")
    parser.add_argument(
            "-prog", dest="prog", metavar="PROG",
            default=".*",
            help="対象とするプロセス名のパターン。(default: 全てのプロセスが対象)")
    parser.add_argument(
            "-args", dest="args", metavar="ARGS",
            help="プロセスに渡されたオプションのパターン。")

    args = parser.parse_args()

    # Docker接続先URLの構築
    args.url = "http://%s:%s" % (args.ip_addr, args.port)

    # 引数を記録
    logger.info("args:%s", args)

    dc = docker.Client(base_url=args.url, version=DOCKER_API_VERSION)
    
    containers = []
    if args.cid :
        pa = re.compile(('%s.*' % args.cid[:len(args.cid) - 1]) if args.cid.endswith('*') else args.cid)
        def find_container_by_id(c):
            if pa.match(c["Id"]):
                containers.append(c["Id"])
        map(find_container_by_id, dc.containers())
    elif args.name :
        pa = re.compile(('%s.*' % args.name[:len(args.name) - 1]) if args.name.endswith('*') else args.name)
        def find_container_by_name(c):
            # 名前に含まれる '/' を除外するため、マッチングする文字列の 1 文字目をスキップ
            matched_names = [n[1:] for n in c["Names"] if pa.match(n[1:])]
            if len(matched_names) > 0:
                containers.append(c["Id"])
        map(find_container_by_name, dc.containers())
    else :
        def all_container(c):
            containers.append(c["Id"])
        map(all_container,dc.containers())
        
    if len(containers) == 0:
        raise LocalException("指定されたコンテナは存在しません。")
    
    retrieve_process_count(dc, args.prog, args.args, containers)

    return 0

#---------------------------------------
# コンテナ上のプロセス数を出力する
#---------------------------------------
def retrieve_process_count(dc, prog, args, containers):
    # プロセスのマッチング用パターン
    pa = re.compile(('%s\s+%s' % (prog, args)) if args else prog)
    
    def print_process_count(cid):
        # コンテナ内のプロセス一覧取得
        r = dc.top(cid)
        # コンテナ数を出力
        print '%s%s:%s,%d' % (prog, ' ' + args if args else '', cid[0:12], len([p for p in r["Processes"] if pa.match(p[r["Titles"].index("CMD")])]))
        
    map(print_process_count, containers)

#===============================================================================
if __name__ == "__main__":
    try:
        sys.exit(main())

    # 予期されている例外はメッセージのみ出力
    except LocalException as e:
        print e.message
        sys.exit(1)

    # 予期しない例外はスタックトレースまで出力
    except Exception as e:
        print e.message
        traceback.print_exc()
        sys.exit(1)
