#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) since 2015 NTT DATA Corporation 
# Last update : 2015/09/15
#
from __future__ import unicode_literals

import argparse
import codecs
import datetime
import dateutil.parser
import docker
import json
import logging.config
import netsnmp
import os
import pprint
import re
import sys
import threading
import time
import traceback
import urllib

SCRIPT_NAME = os.path.splitext(os.path.basename(sys.argv[0]))[0]
DOCKER_API_VERSION = "1.18"
DEFAULT_HOST = "127.0.0.1"
DEFAULT_PORT = "2375"
DEFAULT_SNMP_VER = 1
DEFAULT_SNMP_COM = "public"
DEFAULT_SNMP_PORT = "161"
TMP_DIR = "/tmp"
TMP_FINGERPRINT = SCRIPT_NAME + "_3"
TMP_EXPIRES = 0 # disable
CID_LEN = 8

# 標準入出力をUTF-8で行う
sys.stdin = codecs.getreader("utf-8")(sys.stdin)
sys.stdout = codecs.getwriter("utf-8")(sys.stdout)

# カレントディレクトリをスクリプトのディレクトリへ移動
# loggingを初期化(logging.iniファイルがある場合)
os.chdir(os.path.dirname(os.path.abspath(sys.argv[0])))
if os.path.exists("logging.ini"): logging.config.fileConfig("logging.ini")
logger = logging.getLogger(SCRIPT_NAME)

class LocalException(Exception):
    pass

# 開発/デバッグ用
def dumpvar(var):
    print pprint.pformat(var).decode("utf-8")

#---------------------------------------
# Docker-stats poller
#---------------------------------------
class StatsPoller(threading.Thread):

    def __init__(self, url, cid):
        super(StatsPoller, self).__init__()
        self.__url = url
        self.__cid = cid
        self.container_id = cid # 外部からの参照用

    def run(self):
        c = docker.Client(base_url=self.__url, version=DOCKER_API_VERSION)
        for stats_json in c.stats(self.__cid):
            logger.debug("StatsPoller; stats:%s", stats_json.rstrip())
            self.result = json.loads(stats_json);
            break

#---------------------------------------
# SNMP poller
#---------------------------------------
class SnmpPoller(threading.Thread):

    def __init__(self, host, ver, com, port, *oidlist):
        super(SnmpPoller, self).__init__()
        self.__host = host
        self.__ver = ver
        self.__com = com
        self.__port = port
        self.__oidlist = oidlist

    def run(self):
        res = netsnmp.snmpget(
                *self.__oidlist,
                Version=self.__ver,
                DestHost="%s:%s" % (self.__host, self.__port), # RemotePortパラメータは無視される？
                Community=self.__com)
        self.result = res
        logger.debug("SnmpPoller; host:%s ver:%s com:%s port:%s oids:%s result:%s",
                     self.__host, self.__ver, self.__com, self.__port, str(self.__oidlist), str(res))

#---------------------------------------
# ISO8601文字列をミリ秒に変換
#---------------------------------------
def conv_iso_msec(iso):
    # datetimeへ変換
    dt = dateutil.parser.parse(iso)
    # タイムゾーン情報を除去
    dt = dt.replace(tzinfo=None)
    # ミリ秒へ変換
    msec = time.mktime(dt.timetuple()) * 1000 + float(dt.microsecond) / 1000
    logger.debug("conv_iso_msec; %s -> %s -> %f" % (iso, dt, msec))
    return msec

#---------------------------------------
# メイン
#---------------------------------------
def main():
    logger.info(SCRIPT_NAME)

    # 引数解析
    parser = argparse.ArgumentParser(
            formatter_class=argparse.RawTextHelpFormatter,
            description="DockerからCPU使用率を取得します。")

    parser.add_argument(
            "-i", dest="ip_addr", metavar="IP_ADDRESS",
            default=DEFAULT_HOST,
            help="Docker Remote APIの接続先IPアドレス。(default: '%s')" % DEFAULT_HOST)
    parser.add_argument(
            "-p", dest="port", metavar="DOCKER_PORT",
            default=DEFAULT_PORT,
            help="Docker Remote APIの接続先ポート。(default: '%s')" % DEFAULT_PORT)
    parser.add_argument(
            "-sv", dest="snmp_ver", metavar="SNMP_VER", type=int,
            default=DEFAULT_SNMP_VER,
            help="SNMPのバージョン。(default: '%s')" % DEFAULT_SNMP_VER)
    parser.add_argument(
            "-sc", dest="snmp_com", metavar="SNMP_COM",
            default=DEFAULT_SNMP_COM,
            help="SNMPのコミュニティ。(default: '%s')" % DEFAULT_SNMP_COM)
    parser.add_argument(
            "-sp", dest="snmp_port", metavar="SNMP_PORT",
            default=DEFAULT_SNMP_PORT,
            help="SNMPのポート。(default: '%s')" % DEFAULT_SNMP_PORT)
    group = parser.add_mutually_exclusive_group()
    group.add_argument(
            "-c", dest="cid", metavar="CONTAINER_ID", 
            help="コンテナのIDパターンを指定します。")
    group.add_argument(
            "-n", dest="name", metavar="CONTAINER_NAME",
            help="コンテナの名称パターンを指定します。")
    parser.add_argument(
            "-k", dest="kid", metavar="KEY_ID",
            help="監視ごとにユニークなIDを指定します。")
    parser.add_argument(
            "--all", dest="dummy", action="store_true",
            help="CPU全体の結果を表示します。(--coreとは排他)")
    parser.add_argument(
            "--core", dest="per_cpu", action="store_true",
            help="CPUコア別に結果を表示します。(--allとは排他)")
    parser.add_argument(
            "--with-host", dest="with_host", action="store_true",
            help="ホストのCPU使用率を合わせて表示します。")

    args = parser.parse_args()

    # keyは必須
    if args.kid is None:
        raise LocalException("KEY_IDを指定してください。")

    # Docker接続先URLの構築
    args.url = "http://%s:%s" % (args.ip_addr, args.port)

    # 一時ファイル名の構築
    args.tmpfile = "%s/%s.txt" % (TMP_DIR, urllib.quote(args.kid, safe=""))

    # --with-host は -c, --core オプションとは併用不可
    if args.with_host and (args.cid or args.name or args.per_cpu):
        raise LocalException("'--with-host'オプションは'-c','-n','--core'オプションとは併用できません。")

    # experimental:
    #args.uses_name = False

    # コンテナIDのリストを作成する
    dc = docker.Client(base_url=args.url, version=DOCKER_API_VERSION)
    cid2name = {}
    containers = []
    if args.cid :
        pa = re.compile(('%s.*' % args.cid[:len(args.cid) - 1]) if args.cid.endswith('*') else args.cid)
        def find_container_by_id(c):
            if pa.match(c["Id"]):
                containers.append(c["Id"])
                cid2name[c["Id"]] = c["Names"][0][1:]
        map(find_container_by_id, dc.containers())
    elif args.name :
        pa = re.compile(('%s.*' % args.name[:len(args.name) - 1]) if args.name.endswith('*') else args.name)
        def find_container_by_name(c):
            # 名前に含まれる '/' を除外するため、マッチングする文字列の 1 文字目をスキップ
            matched_names = [n[1:] for n in c["Names"] if pa.match(n[1:])]
            if len(matched_names) > 0:
                containers.append(c["Id"])
                cid2name[c["Id"]] = c["Names"][0][1:]
        map(find_container_by_name, dc.containers())
    else :
        def all_container(c):
            containers.append(c["Id"])
            cid2name[c["Id"]] = c["Names"][0][1:]
        map(all_container,dc.containers())
        
    if len(containers) == 0:
        raise LocalException("指定されたコンテナは存在しません。")

    # 引数を記録
    logger.info("args:%s", args)

    # ポーリング実行
    polling_time = time.time()
    pollers = []
    for cid in containers:
        pollers.append(StatsPoller(args.url, cid))

    # ホストのCPU使用率をポーリングするかどうか
    snmp_poller = None
    if args.with_host:
        snmp_poller = SnmpPoller(
                args.ip_addr, args.snmp_ver, args.snmp_com, args.snmp_port,
                ".1.3.6.1.4.1.2021.11.50.0", #ssCpuRawUser
                ".1.3.6.1.4.1.2021.11.51.0", #ssCpuRawNice
                ".1.3.6.1.4.1.2021.11.52.0", #ssCpuRawSystem
                ".1.3.6.1.4.1.2021.11.53.0", #ssCpuRawIdle
                ".1.3.6.1.4.1.2021.11.54.0") #ssCpuRawWait

    for p in pollers: p.start()
    if snmp_poller: snmp_poller.start()

    for p in pollers: p.join()
    if snmp_poller: snmp_poller.join()

    # 今回値算出
    curr = {
        "fingerprint": TMP_FINGERPRINT,
        "time": polling_time,
        "time_str": time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(polling_time)),
        "containers": {},
    }
    for p in pollers:
        curr["containers"][p.container_id] = {
            "read" : p.result["read"],
            "cpu_stats" : p.result["cpu_stats"]
        }

    if snmp_poller:
        r = []
        for it in snmp_poller.result: r.append(int(it))
        curr["host"] = {
            "ssCpuRawUser": r[0],
            "ssCpuRawNice": r[1],
            "ssCpuRawSystem": r[2],
            "ssCpuRawIdle": r[3],
            "ssCpuRawWait": r[4],
            "ssCpuRawAll": sum(r)
        }

    # 前回値があれば、計算して結果を出力する
    if os.path.exists(args.tmpfile):
        try:
            with open(args.tmpfile) as f: prev = json.load(f)
        except Exception as e:
            logger.warn("JSON decode error: %s", e)
            prev = None

        if prev is None:
            pass
        elif prev.get("fingerprint") != TMP_FINGERPRINT:
            logger.info("Fingerprint mismatch %s != %s", prev.get("fingerprint"), TMP_FINGERPRINT)
        else:
            time_delta = curr["time"] - prev["time"]
            if TMP_EXPIRES > 0 and time_delta > TMP_EXPIRES:
                logger.info("前回値破棄 time_delta:%s", time_delta)
            else:
                calculate_and_output(args, curr, prev, cid2name)

    # 今回値保存
    with open(args.tmpfile, "w") as f: json.dump(curr, f, indent=1)

# 計算～結果出力
# (mainにあると長くてインデント対応が分かりにくいので別関数化)
def calculate_and_output(args, curr, prev, cid2name):
    output_lines = []
    sum_container_values = 0.0

    # 各コンテナ値
    for cid in sorted(curr["containers"].iterkeys()):
        logger.debug("cid:%s", cid)
        curr_it = curr["containers"][cid]
        prev_it = prev["containers"].get(cid)
        if prev_it is None:
            logger.info("前回値なし")
            continue

        curr_deno = conv_iso_msec(curr_it["read"])
        prev_deno = conv_iso_msec(prev_it["read"])
        deno = (curr_deno - prev_deno) * 1000000 # millsec -> nanosec
        num_cpu = len(curr_it["cpu_stats"]["cpu_usage"]["percpu_usage"])

        if args.per_cpu:
            # コア別
            for i in range(0, num_cpu):
                logger.debug("core:%d", i)
                curr_nume = curr_it["cpu_stats"]["cpu_usage"]["percpu_usage"][i]
                prev_nume = prev_it["cpu_stats"]["cpu_usage"]["percpu_usage"][i]
                value = calc_percent(curr_nume - prev_nume, deno)

                output_lines.append({
                    "core": "#%d" % i,
                    "name": cid2name[cid],
                    "value": value
                })
        else:
            # 全体
            curr_nume = curr_it["cpu_stats"]["cpu_usage"]["total_usage"]
            prev_nume = prev_it["cpu_stats"]["cpu_usage"]["total_usage"]
            value = calc_percent(curr_nume - prev_nume, deno)

            if args.with_host:
                value /= num_cpu
                sum_container_values += value
                logger.debug("num_cpu:%s, value:%f", num_cpu, value)

            output_lines.append({
                "core": "",
                "name": cid2name[cid],
                "value": value
            })

    # ホスト値
    if args.with_host and prev.get("host"):
        curr_nume = curr["host"]["ssCpuRawIdle"]
        prev_nume = prev["host"]["ssCpuRawIdle"]
        curr_deno = curr["host"]["ssCpuRawAll"]
        prev_deno = prev["host"]["ssCpuRawAll"]
        value = calc_percent(curr_nume - prev_nume, curr_deno - prev_deno)
        value = 100 - value
        logger.debug("host cpu:%f", value)

        value -= sum_container_values
        if value < 0:
            logger.info("Minus:%f value=0", value)
            value = 0

        output_lines.append({
            "core": "",
            "name": "host",
            "value": value
        })

    # 結果出力
    for it in output_lines:
        # if args.cid is not None:
            #print "cpu_usage%s,%.2f" % (it["core"], it["value"])
        #else:
        print "cpu_usage%s:%s,%.2f" % (it["core"], it["name"], it["value"])

# パーセンテージ計算サブ
def calc_percent(nume, deno):
    # ゼロ除算になる場合は0とする
    if deno == 0:
        logger.warn("calc; ZeroDiv value=0")
        return 0

    # 計算
    value = float(nume) / deno * 100
    logger.debug("calc; %f / %f * 100 = %f", nume, deno, value)

    # 計算結果がマイナスの場合は0とする
    if value < 0:
        logger.info("calc; Minus:%f value=0", value)
        value = 0

    return value

#===============================================================================
if __name__ == "__main__":
    try:
        main()
        sys.exit(0)

    # 予期されている例外はメッセージのみ出力
    except LocalException as e:
        print e.message
        sys.exit(1)

    # 予期しない例外はスタックトレースまで出力
    except Exception as e:
        print e.message
        traceback.print_exc()
        sys.exit(1)
