#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) since 2015 NTT DATA Corporation 
# Last update : 2015/09/15
#
from __future__ import unicode_literals

import argparse
import codecs
import docker
import json
import logging.config
import os
import re
import sys
import threading
import time
import traceback
import urllib

SCRIPT_NAME = os.path.splitext(os.path.basename(sys.argv[0]))[0]
DOCKER_API_VERSION = "1.18"
DEFAULT_HOST = "127.0.0.1"
DEFAULT_PORT = "2375"
TMP_DIR = "/tmp"
TMP_FINGERPRINT = SCRIPT_NAME + "_2"
TMP_EXPIRES = 0 # disable
CID_LEN = 8

# 標準入出力をUTF-8で行う
sys.stdin = codecs.getreader("utf-8")(sys.stdin)
sys.stdout = codecs.getwriter("utf-8")(sys.stdout)

# カレントディレクトリをスクリプトのディレクトリへ移動
# loggingを初期化(logging.iniファイルがある場合)
os.chdir(os.path.dirname(os.path.abspath(sys.argv[0])))
if os.path.exists("logging.ini"): logging.config.fileConfig("logging.ini")
logger = logging.getLogger(SCRIPT_NAME)

class LocalException(Exception):
    pass

#---------------------------------------
# Docker-stats poller
#---------------------------------------
class StatsPoller(threading.Thread):

    def __init__(self, url, cid):
        super(StatsPoller, self).__init__()
        self.__url = url
        self.__cid = cid
        self.container_id = cid # 外部からの参照用

    def run(self):
        c = docker.Client(base_url=self.__url, version=DOCKER_API_VERSION)
        for stats_json in c.stats(self.__cid):
            logger.debug("StatsPoller; stats:%s", stats_json.rstrip())
            self.result = json.loads(stats_json);
            break

#---------------------------------------
# メイン
#---------------------------------------
def main():
    logger.info(SCRIPT_NAME)

    # 引数解析
    parser = argparse.ArgumentParser(
            formatter_class=argparse.RawTextHelpFormatter,
            description="DockerのディスクI/O情報を取得します。")

    parser.add_argument(
            "-i", dest="ip_addr", metavar="IP_ADDRESS",
            default=DEFAULT_HOST,
            help="Docker Remote APIの接続先IPアドレス。(default: '%s')" % DEFAULT_HOST)
    parser.add_argument(
            "-p", dest="port", metavar="DOCKER_PORT",
            default=DEFAULT_PORT,
            help="Docker Remote APIの接続先ポート。(default: '%s')" % DEFAULT_PORT)
    group = parser.add_mutually_exclusive_group()
    group.add_argument(
            "-c", dest="cid", metavar="CONTAINER_ID", 
            help="コンテナのIDパターンを指定します。")
    group.add_argument(
            "-n", dest="name", metavar="CONTAINER_NAME",
            help="コンテナの名称パターンを指定します。")
    parser.add_argument(
            "-k", dest="kid", metavar="KEY_ID",
            help="監視ごとにユニークなIDを指定します。")
    parser.add_argument(
            "--read", dest="read_flag", action="store_true",
            help="read値のみ出力します。")
    parser.add_argument(
            "--write", dest="write_flag", action="store_true",
            help="write値のみ出力します。")

    args = parser.parse_args()

    # keyは必須
    if args.kid is None:
        raise LocalException("KEY_IDを指定してください。")

    # read/write省略時は双方有効
    if not args.read_flag and not args.write_flag:
        args.read_flag = True
        args.write_flag = True

    # Docker接続先URLの構築
    args.url = "http://%s:%s" % (args.ip_addr, args.port)

    # 一時ファイル名の構築
    args.tmpfile = "%s/%s.txt" % (TMP_DIR, urllib.quote(args.kid, safe=""))

    # コンテナIDのリストを作成する
    dc = docker.Client(base_url=args.url, version=DOCKER_API_VERSION)
    
    containers = []
    if args.cid :
        pa = re.compile(('%s.*' % args.cid[:len(args.cid) - 1]) if args.cid.endswith('*') else args.cid)
        def find_container_by_id(c):
            if pa.match(c["Id"]):
                containers.append(c["Id"])
        map(find_container_by_id, dc.containers(quiet=True))
    elif args.name :
        pa = re.compile(('%s.*' % args.name[:len(args.name) - 1]) if args.name.endswith('*') else args.name)
        def find_container_by_name(c):
            # 名前に含まれる '/' を除外するため、マッチングする文字列の 1 文字目をスキップ
            matched_names = [n[1:] for n in c["Names"] if pa.match(n[1:])]
            if len(matched_names) > 0:
                containers.append(c["Id"])
        map(find_container_by_name, dc.containers())
    else :
        def all_container(c):
            containers.append(c["Id"])
        map(all_container,dc.containers(quiet=True))
        
    if len(containers) == 0:
        raise LocalException("指定されたコンテナは存在しません。")

    # 引数を記録
    logger.info("args:%s", args)

    # ポーリング実行
    polling_time = time.time()
    pollers = []
    for cid in containers:
        pollers.append(StatsPoller(args.url, cid))
    for p in pollers: p.start()
    for p in pollers: p.join()

    # 今回値算出
    curr = {
        "fingerprint": TMP_FINGERPRINT,
        "time": polling_time,
        "time_str": time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(polling_time)),
        "containers": {}
    }
    for p in pollers:
        sum = {
            "read":  0,
            "write": 0
        }
        #print p.container_id
        curr["containers"][p.container_id] = sum
        for io in p.result["blkio_stats"]["io_service_bytes_recursive"]:
            if io["op"] == "Read":
                sum["read"] += io["value"]
            elif io["op"] == "Write":
                sum["write"] += io["value"]

    # 前回値があれば、計算して結果を出力する
    if os.path.exists(args.tmpfile):
        try:
            with open(args.tmpfile) as f: prev = json.load(f)
        except Exception as e:
            logger.warn("JSON decode error: %s", e)
            prev = {}

        if prev.get("fingerprint") != TMP_FINGERPRINT:
            logger.info("Fingerprint mismatch %s != %s", prev.get("fingerprint"), TMP_FINGERPRINT)
        else:
            # 結果出力
            time_delta = curr["time"] - prev["time"]
            if TMP_EXPIRES > 0 and time_delta > TMP_EXPIRES:
                logger.info("前回値破棄 time_delta:%s", time_delta)
            else:
                logger.debug("time_d:%s", time_delta)
                if time_delta == 0: # zero-div対策
                    logger.info("time_delta 0 -> 1.")
                    time_delta = 1

                for cid in sorted(curr["containers"]):
                    logger.debug("cid:%s", cid)
                    curr_it = curr["containers"][cid]
                    prev_it = prev["containers"].get(cid)
                    if prev_it is None:
                        logger.info("前回値なし cid:%s", cid)
                    else:
                        # 計算＆出力
                        def calc(mode, curr_val, prev_val):
                            delta = curr_val - prev_val
                            value = float(delta) / time_delta / 1024
                            if args.cid or args.name :
                                # コンテナ指定の場合はkeyを出力しない
                                print "%s,%.2f" % (mode, value)
                            else:
                                print "%s:%s,%.2f" % (mode, cid[:CID_LEN], value)
                            logger.debug("%s curr:%s prev:%s d:%s val:%s", mode, curr_val, prev_val, delta, value)

                        if args.read_flag:
                            calc("disk_read", curr_it["read"], prev_it["read"])
                        if args.write_flag:
                            calc("disk_write", curr_it["write"], prev_it["write"])
    # 今回値保存
    with open(args.tmpfile, "w") as f: json.dump(curr, f, indent=1)

#===============================================================================
if __name__ == "__main__":
    try:
        main()
        sys.exit(0)

    # 予期されている例外はメッセージのみ出力
    except LocalException as e:
        print e.message
        sys.exit(1)

    # 予期しない例外はスタックトレースまで出力
    except Exception as e:
        print e.message
        traceback.print_exc()
        sys.exit(1)
