#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) since 2015 NTT DATA Corporation 
# Last update : 2015/09/15
#
from __future__ import unicode_literals

import argparse
import codecs
import docker
import json
import logging.config
import os
import socket
import sys
import syslog
import time
import traceback
import urllib

SCRIPT_NAME = os.path.splitext(os.path.basename(sys.argv[0]))[0]
DOCKER_API_VERSION = "1.18"
DEFAULT_HOST = "127.0.0.1"
DEFAULT_PORT = "2375"
TMP_DIR = "/tmp"
TMP_FINGERPRINT = SCRIPT_NAME + "_1"
TMP_EXPIRES = 0 # disable

# 標準入出力をUTF-8で行う
sys.stdin = codecs.getreader("utf-8")(sys.stdin)
sys.stdout = codecs.getwriter("utf-8")(sys.stdout)

# カレントディレクトリをスクリプトのディレクトリへ移動
# loggingを初期化(logging.iniファイルがある場合)
os.chdir(os.path.dirname(os.path.abspath(sys.argv[0])))
if os.path.exists("logging.ini"): logging.config.fileConfig("logging.ini")
logger = logging.getLogger(SCRIPT_NAME)

class LocalException(Exception):
    pass

#---------------------------------------
# メイン
#---------------------------------------
def main():
    logger.info(SCRIPT_NAME)

    # 引数解析
    parser = argparse.ArgumentParser(
            formatter_class=argparse.RawTextHelpFormatter,
            description="Dockerのイベント情報を取得します。")

    parser.add_argument(
            "-i", dest="ip_addr", metavar="IP_ADDRESS",
            default=DEFAULT_HOST,
            help="Docker Remote APIの接続先IPアドレス。(default: '%s')" % DEFAULT_HOST)
    parser.add_argument(
            "-p", dest="port", metavar="DOCKER_PORT",
            default=DEFAULT_PORT,
            help="Docker Remote APIの接続先ポート。(default: '%s')" % DEFAULT_PORT)
    parser.add_argument(
            "-k", dest="kid", metavar="KEY_ID",
            help="監視ごとにユニークなIDを指定します。")
    parser.add_argument(
            "-s", dest="syslog", metavar="SYSLOG_HOST:PORT",
            help="イベントをsyslogとして送信します。")
    parser.add_argument(
            "-f", dest="logfile", metavar="LOGFILE",
            help="イベントをファイルへ出力します。")

    args = parser.parse_args()

    # keyは必須
    if args.kid is None:
        raise LocalException("KEY_IDを指定してください。")

    # -s/-fはいずれか必須
    if not args.syslog and not args.logfile:
        raise LocalException("-sまたは-fオプションのいずれかは必須です。")

    # Docker接続先URLの構築
    args.url = "http://%s:%s" % (args.ip_addr, args.port)

    # 一時ファイル名の構築
    args.tmpfile = "%s/%s.txt" % (TMP_DIR, urllib.quote(args.kid, safe=""))

    # システムログの送信先IPアドレスとポートを分離
    if args.syslog:
        args.syslog_host = args.syslog.split(":")[0]
        args.syslog_port = int(args.syslog.split(":")[1])

    # 引数を記録
    logger.info("args:%s", args)

    # 前回値を取得する
    lasttime = 0  # 使用時に+1するので、ここでは0にしておく
    if os.path.exists(args.tmpfile):
        try:
            with open(args.tmpfile) as f: prev = json.load(f)
        except Exception as e:
            logger.warn("JSON decode error: %s", e)
            prev = {}

        if prev.get("fingerprint") != TMP_FINGERPRINT:
            logger.info("Fingerprint mismatch %s != %s", prev.get("fingerprint"), TMP_FINGERPRINT)
        else:
            lasttime = prev["lasttime"]

    # ポーリング実行
    dc = docker.Client(base_url=args.url, version=DOCKER_API_VERSION)
    now = int(time.time())
    logger.info("since:%s until:%s", lasttime + 1, now)
    for evt in dc.events(since=lasttime + 1, until=now):
        try:
            evt = json.loads(evt)
            lasttime = evt["time"]
            if args.logfile: main_logfile(args, evt)
            if args.syslog: main_syslog(args, evt)
        except Exception as e:
            logger.warn("JSON decode error: %s", e)

    # 今回値算出
    curr = {
        "fingerprint": TMP_FINGERPRINT,
        "lasttime": lasttime,
        "lasttime_str": time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(lasttime)),
    }

    # 今回値保存
    with open(args.tmpfile, "w") as f: json.dump(curr, f, indent=1)

#---------------------------------------
# メイン：ログファイルへ出力
#---------------------------------------
def main_logfile(args, event):
    with open(args.logfile, "a") as f:
        line = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(event["time"]))
        line += ' %s' % (args.ip_addr)
        # "from"はNoneのことがあるのでgetする
        line += ' "status":"%s","id":"%s","from":"%s"' % (event["status"], event["id"], event.get("from"))
        line += "\n"
        f.write(line)

#---------------------------------------
# メイン：syslogへ出力
#---------------------------------------
def main_syslog(args, event):
    # socketを作成して直接送る
    sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)

    # 各パートの内容をセットアップ
    pri = syslog.LOG_INFO + syslog.LOG_USER
    timestamp = get_syslog_timestamp(event["time"])
    hostname = args.ip_addr
    msg = str(event)

    # syslogパケットを構築
    packet = "<%s>%s %s %s" % (pri, timestamp, hostname, msg)
    packet = packet.encode("utf-8") # unicode -> str(utf-8)
    packet = packet[:1024]          # 最大1024バイト

    # パケット送信
    sock.sendto(packet, (args.syslog_host, args.syslog_port))
    logger.debug("syslog packet:%s", packet)

# 時刻(ローカル)をsyslogパケットの書式にする
def get_syslog_timestamp(tm):
    localtime = time.localtime(tm)
    day = time.strftime("%d", localtime)
    if day[0] == "0": day = " " + day[1:]   # "01"～"09" -> " 1"～" 9"
    value = time.strftime("%b %%s %H:%M:%S", localtime) # "%%s" -> "%s"
    return value % day

#===============================================================================
if __name__ == "__main__":
    try:
        main()
        sys.exit(0)

    # 予期されている例外はメッセージのみ出力
    except LocalException as e:
        print e.message
        sys.exit(1)

    # 予期しない例外はスタックトレースまで出力
    except Exception as e:
        print e.message
        traceback.print_exc()
        sys.exit(1)
