#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) since 2015 NTT DATA Corporation 
# Last update : 2015/09/15
#
from __future__ import unicode_literals

import argparse
import codecs
import docker
import logging.config
import os
import sys
import traceback

SCRIPT_NAME = os.path.splitext(os.path.basename(sys.argv[0]))[0]
DOCKER_API_VERSION = "1.18"
DEFAULT_HOST = "127.0.0.1"
DEFAULT_PORT = "2375"

# 標準入出力をUTF-8で行う
sys.stdin = codecs.getreader("utf-8")(sys.stdin)
sys.stdout = codecs.getwriter("utf-8")(sys.stdout)

# カレントディレクトリをスクリプトのディレクトリへ移動
# loggingを初期化(logging.iniファイルがある場合)
os.chdir(os.path.dirname(os.path.abspath(sys.argv[0])))
if os.path.exists("logging.ini"): logging.config.fileConfig("logging.ini")
logger = logging.getLogger(SCRIPT_NAME)

class LocalException(Exception):
    pass

#---------------------------------------
# Dockerコンテナのステータスを、human-readableな表現から単語表現へ変換する
#---------------------------------------
def conv_docker_status(status):
    # 起動エラーによりステータスが空欄となることがある
    if status == "": return "exited"
    w = status.split(" ")
    if w[0] == "Exited": return "exited"
    if w[0] == "Restarting": return "restarting"
    if w[0] == "Up":
        if w[-1] == "(Paused)": return "paused"
        return "running"
    raise LocalException("変換できないステータス:%s" % status)

#---------------------------------------
# メイン
#---------------------------------------
def main():
    logger.info(SCRIPT_NAME)

    # 引数解析
    parser = argparse.ArgumentParser(
            formatter_class=argparse.RawTextHelpFormatter,
            description="Dockerホストの状態を取得します。")

    parser.add_argument(
            "-i", dest="ip_addr", metavar="IP_ADDRESS",
            default=DEFAULT_HOST,
            help="Docker Remote APIの接続先IPアドレス。(default: '%s')" % DEFAULT_HOST)
    parser.add_argument(
            "-p", dest="port", metavar="DOCKER_PORT",
            default=DEFAULT_PORT,
            help="Docker Remote APIの接続先ポート。(default: '%s')" % DEFAULT_PORT)
    parser.add_argument(
            "--status", dest="status", action="store_true",
            help="Dockerホストの正常性(0:正常、1:異常)を返却します。デフォルト動作です。")
    parser.add_argument(
            "--count", dest="count", action="store_true",
            help="Dockerホスト上のステータス別コンテナ数を返却します。")
    parser.add_argument(
            "--size", dest="size", action="store_true",
            help="Dockerホスト上のコンテナが使用するディスク容量を返却します。")

    args = parser.parse_args()

    # Docker接続先URLの構築
    args.url = "http://%s:%s" % (args.ip_addr, args.port)

    # 引数を記録
    logger.info("args:%s", args)

    # 処理分岐
    if args.count:
        main_count(args)
    elif args.size:
        main_size(args)
    else:
        main_status(args)

#---------------------------------------
# メイン：Dockerホストの正常性
#---------------------------------------
def main_status(args):

    # ping実行
    dc = docker.Client(base_url=args.url, version=DOCKER_API_VERSION)
    result = False
    try:
        result = dc.ping()
    except Exception:
        pass    # result = False

    # 結果出力
    if result:
        value = "0"
    else:
        value = "1"
    print "Status,%s" % value

#---------------------------------------
# メイン：コンテナ数
#---------------------------------------
def main_count(args):

    run_cnt = 0 # running
    rst_cnt = 0 # restarting
    pau_cnt = 0 # paused
    ext_cnt = 0 # exited

    dc = docker.Client(base_url=args.url, version=DOCKER_API_VERSION)
    for it in dc.containers(all=True):
        state = conv_docker_status(it["Status"])
        logger.debug("cid:%s state:%s", it["Id"], state)
        if state == "running":
            run_cnt += 1
        elif state == "restarting":
            rst_cnt += 1
        elif state == "paused":
            pau_cnt += 1
        else:
            ext_cnt += 1

    print "running,%s" % run_cnt
    print "restarting,%s" % rst_cnt
    print "paused,%s" % pau_cnt
    print "exited,%s" % ext_cnt

#---------------------------------------
# メイン：ディスク使用容量
#---------------------------------------
def main_size(args):

    img_size = 0 # image
    run_size = 0 # running
    rst_size = 0 # restarting
    pau_size = 0 # paused
    ext_size = 0 # exited

    dc = docker.Client(base_url=args.url, version=DOCKER_API_VERSION, timeout=600)

    # イメージ
    for it in dc.images(all=True):
        size = it["Size"]
        logger.debug("iid:%s size:%d tags:%s", it["Id"], size, str(it["RepoTags"]))
        img_size += size

    # ステータス別コンテナ
    for it in dc.containers(all=True, size=True):
        state = conv_docker_status(it["Status"])
        size = it["SizeRw"]
        logger.debug("cid:%s state:%s size:%d", it["Id"], state, size)
        if state == "running":
            run_size += size
        elif state == "restarting":
            rst_size += size
        elif state == "paused":
            pau_size += size
        else:
            ext_size += size

    # 結果出力
    print "image,%s" % img_size
    print "running,%s" % run_size
    print "restarting,%s" % rst_size
    print "paused,%s" % pau_size
    print "exited,%s" % ext_size

#===============================================================================
if __name__ == "__main__":
    try:
        main()
        sys.exit(0)

    # 予期されている例外はメッセージのみ出力
    except LocalException as e:
        print e.message
        sys.exit(1)

    # 予期しない例外はスタックトレースまで出力
    except Exception as e:
        print e.message
        traceback.print_exc()
        sys.exit(1)
