#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) since 2015 NTT DATA Corporation 
# Last update : 2015/09/15
#
from __future__ import unicode_literals

import argparse
import codecs
import datetime
import docker
import logging.config
import os
import pprint
import suds
import sys
import syslog
import traceback

SCRIPT_NAME = os.path.splitext(os.path.basename(sys.argv[0]))[0]
DOCKER_API_VERSION = "1.18"
DEFAULT_HOST = "127.0.0.1"
DEFAULT_PORT = "2375"
DEFAULT_HINEMOS_URL = "http://127.0.0.1:8080/HinemosWS"
SCOPE_CONTAINERS = "docker_container"
SCOPE_HOSTS = "docker_host"
SCOPE_MAP = "docker_map"
OWNER_ROLE = "ALL_USERS"
CID_LEN = 12

# 標準入出力をUTF-8で行う
sys.stdin = codecs.getreader("utf-8")(sys.stdin)
sys.stdout = codecs.getwriter("utf-8")(sys.stdout)

# カレントディレクトリをスクリプトのディレクトリへ移動
# loggingを初期化(logging.iniファイルがある場合)
os.chdir(os.path.dirname(os.path.abspath(sys.argv[0])))
if os.path.exists("logging.ini"): logging.config.fileConfig("logging.ini")
logger = logging.getLogger(SCRIPT_NAME)

class LocalException(Exception):
    pass

# 開発/デバッグ用
def dumpvar(var):
    print pprint.pformat(var).decode("utf-8")

#---------------------------------------
# メッセージ出力
#---------------------------------------
def trace(message, to_syslog=False):
    # syslog
    if to_syslog: syslog.syslog(syslog.LOG_INFO, message.encode("utf-8"))
    # log
    logger.debug(message)
    # console
    print message

#---------------------------------------
# mapスコープ配下のスコープ名を返す
#---------------------------------------
def get_map_name(host_id):
    return "%s_scope" % host_id

#---------------------------------------
# Dockerコンテナのステータスを、human-readableな表現から単語表現へ変換する
#---------------------------------------
def conv_docker_status(status):
    # 起動エラーによりステータスが空欄となることがある
    if status == "": return "exited"
    w = status.split(" ")
    if w[0] == "Exited": return "exited"
    if w[0] == "Restarting": return "restarting"
    if w[0] == "Up":
        if w[-1] == "(Paused)": return "paused"
        return "running"
    raise LocalException("変換できないステータス:%s" % status)

#---------------------------------------
# メイン
#---------------------------------------
def main():
    logger.info(SCRIPT_NAME)

    # 引数解析
    parser = argparse.ArgumentParser(
            formatter_class=argparse.RawTextHelpFormatter,
            description="Dockerノードを自動登録します。")

    parser.add_argument(
            "-i", dest="ip_addr", metavar="IP_ADDRESS",
            default=DEFAULT_HOST,
            help="Docker Remote APIの接続先IPアドレス。(default: '%s')" % DEFAULT_HOST)
    parser.add_argument(
            "-p", dest="port", metavar="DOCKER_PORT",
            default=DEFAULT_PORT,
            help="Docker Remote APIの接続先ポート。(default: '%s')" % DEFAULT_PORT)
    parser.add_argument(
            "-m", dest="hinemos_url", metavar="HINEMOS_URL",
            default=DEFAULT_HINEMOS_URL,
            help="HinemosマネージャのURL。(default: '%s')" % DEFAULT_HINEMOS_URL)
    parser.add_argument(
            "-f", dest="facility_id", metavar="HOST_FACILITY_ID",
            help="Dockerホストに対応するノードのファシリティID。")

    args = parser.parse_args()

    # ファシリティIDは必須
    if args.facility_id is None:
        raise LocalException("ファシリティIDの指定(-f)は必須です。")

    # Docker接続先URLの構築
    args.url = "http://%s:%s" % (args.ip_addr, args.port)

    # Hinemos URLの末尾に"/"がある場合は除去
    if args.hinemos_url.endswith("/"):
        args.hinemos_url = args.hinemos_url[:-1]

    # Hinemosへのログイン情報
    args.hinemos_username = "hinemos"
    args.hinemos_password = "hinemos"

    # 引数を記録
    logger.info("args:%s", args)

    # 主処理
    syslog.openlog(str("RepositoryDockerNodeAdd"))
    trace("開始しました。", True)
    try:
        hclient = login_hinemos(args.hinemos_url, args.hinemos_username, args.hinemos_password)

        dclient = docker.Client(base_url=args.url, version=DOCKER_API_VERSION)
        containers = get_containers_info(dclient, args.facility_id)

        register_docker_scopes(hclient, args.facility_id)
        register_host_node(hclient, args.facility_id, args.ip_addr)
        assign_host_node(hclient, args.facility_id)
        register_container_nodes(hclient, args.facility_id, args.ip_addr, containers)
        assign_container_nodes(hclient, SCOPE_CONTAINERS, containers)
        assign_container_nodes(hclient, get_map_name(args.facility_id), containers)
        assign_container_label(hclient, containers)

        trace("完了しました。", True)

    except:
        exc_info = sys.exc_info()
        trace("Error: %s,%s" % (exc_info[0], exc_info[1]), True)
        raise

    finally:
        syslog.closelog()

#---------------------------------------
# Hinemosマネージャへログイン
#---------------------------------------
def login_hinemos(url, username, password):
    trace("Hinemosマネージャにログインしています。")

    endpoint = "%s/RepositoryEndpoint?wsdl" % url

    trace("  endpoint: %s" % endpoint)
    trace("  username: %s" % username)

    hclient = suds.client.Client(
            endpoint,
            transport=suds.transport.http.HttpAuthenticated(
                username=username, password=password))

    return hclient

#---------------------------------------
# Docker コンテナ情報を取得する
#---------------------------------------
def get_containers_info(dclient, host_id):
    trace("Dockerコンテナの情報を取得します。")

    containers = dclient.containers(all=True)

    # 後から参照する値をここで算出し、dict へ追加設定する(prefix = "+")
    for it in containers:

        # inspect情報
        it["+inspect"] = dclient.inspect_container(it["Id"])

        # ファシリティID
        it["+facility_id"] = "%s-%s" % (host_id, it["Names"][0][1:])

        # ステータス名
        it["+status_name"] = conv_docker_status(it["Status"])

        # ポート情報
        ports = []
        for port in it["Ports"]:
            try:
                ports.append("%s:%s->%s/%s" % (
                        port["IP"], port["PublicPort"], port["PrivatePort"], port["Type"]))
            except KeyError:
                print ("port not bind")
        it["+ports"] = ", ".join(ports)

        trace("  Id:%s Name:%s Status:'%s':%s" % (
                it["Id"][:CID_LEN], it["Names"][0][1:], it["Status"], it["+status_name"]))

    return containers

#---------------------------------------
# Docker専用スコープを登録する
#---------------------------------------
def register_docker_scopes(hclient, facility_id):
    trace("スコープを作成します。")

    # ルート直下のスコープ
    register_scopes(hclient,
                    parent=None,
                    ids=[SCOPE_CONTAINERS, SCOPE_HOSTS, SCOPE_MAP])

    # map配下のホスト別スコープ
    register_scopes(hclient,
                    parent=SCOPE_MAP,
                    ids=[get_map_name(facility_id)])

# スコープ登録サブ
def register_scopes(hclient, parent, ids):
    present_ids = hclient.service.getFacilityIdList(parent, 1)   #1:直下

    for facility_id in ids:
        if facility_id in present_ids:
            trace("  [skip] %s は既に存在します。" % facility_id)
            continue

        trace("  %s を作成します。" % facility_id)

        scope_info = hclient.factory.create("scopeInfo")
        scope_info.facilityId = facility_id
        scope_info.facilityName = facility_id
        scope_info.description = "スクリプトによる自動登録"
        scope_info.ownerRoleId = OWNER_ROLE
        scope_info.iconImage = ""
        hclient.service.addScope(parent, scope_info)

#---------------------------------------
# Dockerホストノードを登録する
#---------------------------------------
def register_host_node(hclient, facility_id, ip_address):
    trace("Dockerホストノードを作成します。")

    present_ids = hclient.service.getFacilityIdList(facility_id, 1)
    if facility_id in present_ids:
        trace("  [skip] %s は既に存在します。" % facility_id)
        return

    trace("  %s を作成します。" % facility_id)

    node_info = hclient.factory.create("nodeInfo")
    node_info.facilityId = facility_id
    node_info.facilityName = facility_id
    node_info.description = "Auto detect at %s" % datetime.datetime.today()
    node_info.valid = True
    node_info.autoDeviceSearch = 1  # 有効
    node_info.platformFamily = "LINUX"
    node_info.subPlatformFamily = ""
    node_info.ipAddressVersion = 4
    node_info.ipAddressV4 = ip_address
    node_info.nodeName = facility_id
    node_info.ownerRoleId = OWNER_ROLE

    hostname_info = hclient.factory.create("nodeHostnameInfo")
    hostname_info.hostname = facility_id
    node_info.nodeHostnameInfo = [hostname_info]

    hclient.service.addNode(node_info)

#---------------------------------------
# Dockerホストノードをスコープへ割り当てる
#---------------------------------------
def assign_host_node(hclient, facility_id):
    trace("Dockerホストノードを%sスコープへ割当てます。" % SCOPE_HOSTS)

    present_ids = hclient.service.getFacilityIdList(SCOPE_HOSTS, 1)
    if facility_id in present_ids:
        trace("  [skip] %s は既に割当てられています。" % facility_id)
        return

    trace("  %s を割当てます。" % facility_id)
    hclient.service.assignNodeScope(SCOPE_HOSTS, [facility_id])

#---------------------------------------
# Dockerコンテナノードを登録する
#---------------------------------------
def register_container_nodes(hclient, host_id, host_address, containers):
    trace("Dockerコンテナノードを作成します。")

    for container in containers:
        facility_id = container["+facility_id"]

        present_ids = hclient.service.getFacilityIdList(facility_id, 1)
        if facility_id in present_ids:
            trace("  [skip] %s は既に存在します。" % facility_id)
            continue

        trace("  %s を作成します。" % facility_id)

        node_info = hclient.factory.create("nodeInfo")
        node_info.facilityId = facility_id
        node_info.facilityName = container["Names"][0][1:]
        node_info.description = "Auto detect at %s" % datetime.datetime.today()
        node_info.valid = (container["+status_name"] == "running")
        node_info.autoDeviceSearch = 0  # 無効
        node_info.platformFamily = "LINUX"
        node_info.subPlatformFamily = ""
        node_info.ipAddressVersion = 4
        node_info.ipAddressV4 = host_address
        node_info.nodeName = container["+inspect"]["Config"]["Hostname"]
        node_info.ownerRoleId = OWNER_ROLE

        hostname_info = hclient.factory.create("nodeHostnameInfo")
        hostname_info.hostname = container["+inspect"]["Config"]["Hostname"]
        node_info.nodeHostnameInfo = [hostname_info]

        # ノード変数
        def node_var(name, value):
            var_info = hclient.factory.create("nodeVariableInfo")
            var_info.nodeVariableName = name
            var_info.nodeVariableValue = value
            return var_info

        node_info.nodeVariableInfo = [
            node_var("DOCKER_HOST", host_id),
            node_var("DOCKER_CONTAINER_ID", container["Id"][:CID_LEN]),
            node_var("DOCKER_COMMAND", container["Command"]),
            node_var("DOCKER_PORTS", container["+ports"]),
            node_var("DOCKER_IMAGE", container["Image"]),
            node_var("DOCKER_STATUS", container["+status_name"]),
        ]

        hclient.service.addNode(node_info)

#---------------------------------------
# Dockerコンテナノードをスコープへ割り当てる
#---------------------------------------
def assign_container_nodes(hclient, scope_id, containers):
    trace("Dockerコンテナノードを%sスコープへ割当てます。" % scope_id)

    present_ids = hclient.service.getFacilityIdList(scope_id, 1)

    for container in containers:
        facility_id = container["+facility_id"]

        if facility_id in present_ids:
            trace("  [skip] %s は既に割当てられています。" % facility_id)
            continue

        trace("  %s を割当てます。" % facility_id)
        hclient.service.assignNodeScope(scope_id, [facility_id])

#---------------------------------------
# Dockerコンテナノードをラベル定義されているスコープへ割り当てる
#---------------------------------------
def assign_container_label(hclient, containers):
    
    exit_scope_ids = hclient.service.getFacilityIdList(None, 1)
    idlist={}
    for exit_scope_id in exit_scope_ids :
        idlist[exit_scope_id]=hclient.service.getFacilityIdList(exit_scope_id, 1)

    for container in containers :
        facility_id = container["+facility_id"]
        scope_ids = container["Labels"]["hinemosAssignScopeId"].split(",") if container["Labels"].has_key("hinemosAssignScopeId") else []
        
        if len(scope_ids) is 0 :
            continue

        trace("%s をラベル設定からスコープへ割当てます。対象スコープ： %s" % (facility_id, scope_ids))
        for label_id in scope_ids :
            if label_id in exit_scope_ids :
                if facility_id in idlist[label_id] :
                    trace("  [skip] %s は既に割当てられています。" % facility_id)
                else :
                    trace("  %s を割当てます。" % facility_id)
                    hclient.service.assignNodeScope(label_id, [facility_id])
            else :
                trace("  [skip] スコープ %s が存在していないため、割当てをスキップします。" % label_id)

#===============================================================================
if __name__ == "__main__":
    try:
        exit_code = main()
        sys.exit(exit_code)

    # 予期されている例外はメッセージのみ出力
    except LocalException as e:
        print e.message
        sys.exit(1)

    # 予期しない例外はスタックトレースまで出力
    except Exception as e:
        print e.message
        traceback.print_exc()
        sys.exit(1)
